/****************************************************************************
    MP4 input module

    Copyright (C) 2017 Krzysztof Nikiel

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************************/

/* 2022.03.19 AP: modified to work with POCC and multithreading */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <limits.h>

#ifdef __POCC__
 #define faad_fopen fopen
#else
 #include "unicode_support.h"
#endif
#include "mp4read.h"
#include "aac_types.h"

enum ATOM_TYPE
{
    ATOM_STOP = 0, /* end of atoms */
    ATOM_NAME,     /* plain atom */
    ATOM_DESCENT,               /* starts group of children */
    ATOM_ASCENT,                /* ends group */
    ATOM_DATA,
    ATOM_FORMAT,   /* format atom; skip if not found */
    ATOM_DESCRIPT
};

enum
{
    ERR_OK = 0,
    ERR_FAIL = -1,
    ERR_MEMORY = -2,
    ERR_UNSUPPORTED = -4,
    ERR_CODEC_UNSUPPORTED = -8
};

typedef struct
{
    uint16_t opcode;
    void *data;
} creator_t;

typedef int (*fileAtomIO)(FILE *,
        int size,
        mp4config_t*,
        creator_t **atom);

//mp4config_t mp4config = { 0 };

/* macros */

static inline uint32_t bswap32(const uint32_t u32)
{
#ifndef WORDS_BIGENDIAN
#if defined (__GNUC__) && ((__GNUC__ > 4) || ((__GNUC__ == 4) && (__GNUC_MINOR__ >= 3)))
    return __builtin_bswap32(u32);
#elif defined (_MSC_VER)
 #ifdef __POCC__
  #pragma intrinsic(_bswap)
    return _bswap(u32);
 #else
    return _byteswap_ulong(u32);
 #endif
#else
    return (u32 << 24) | ((u32 << 8) & 0xFF0000) | ((u32 >> 8) & 0xFF00) | (u32 >> 24);
#endif
#else
    return u32;
#endif
}

static inline uint16_t bswap16(const uint16_t u16)
{
#ifndef WORDS_BIGENDIAN
#if defined (__GNUC__) && ((__GNUC__ > 4) || ((__GNUC__ == 4) && (__GNUC_MINOR__ >= 8)))
    return __builtin_bswap16(u16);
#elif defined (_MSC_VER) && !defined(__POCC__)
    return _byteswap_ushort(u16);
#else
    return (u16 << 8) | (u16 >> 8);
#endif
#else
    return u16;
#endif
}

/* code */

static int datain(FILE *f, void *data, int size)
{
    if (fread(data, 1, size, f) != size)
        return ERR_FAIL;
    return size;
}

static int stringin(FILE *f, char *txt, int sizemax)
{
    int size;
    for (size = 0; size < sizemax; size++)
    {
        if (fread(txt + size, 1, 1, f) != 1)
            return ERR_FAIL;
        if (!txt[size])
            break;
    }
    txt[sizemax-1] = '\0';

    return size;
}

static uint32_t u32in(FILE *f)
{
    uint32_t u32;
    datain(f, &u32, 4);
    u32 = bswap32(u32);
    return u32;
}

static uint16_t u16in(FILE *f)
{
    uint16_t u16;
    datain(f, &u16, 2);
    u16 = bswap16(u16);
    return u16;
}

static int u8in(FILE *f)
{
    uint8_t u8;
    datain(f, &u8, 1);
    return u8;
}


static uint32_t getsize(FILE *f)
{
    int cnt;
    uint32_t size = 0;
    for (cnt = 0; cnt < 4; cnt++)
    {
        int tmp = u8in(f);

        size <<= 7;
        size |= (tmp & 0x7f);
        if (!(tmp & 0x80))
            break;
    }
    return size;
}

#ifndef NO_MP4_HEADER_INFO
enum
{ SECSINDAY = 24 * 60 * 60 };
static char *mp4time(time_t t)
{
    int y;

    // subtract some seconds from the start of 1904 to the start of 1970
    for (y = 1904; y < 1970; y++)
    {
        t -= 365 * SECSINDAY;
        if (!(y & 3))
            t -= SECSINDAY;
    }
    return ctime(&t);
}
#endif

static int mdatin(FILE *f, int size, mp4config_t *config)
{
    if (!config->ftyp)
        config->ftyp = 'qt  ';
    if (!config->dataofs)
        config->dataofs = ftell(f);
    //if (!config->datasize)
    //    config->datasize = size;
    return size;
}

static int ftypin(FILE *f, int size, mp4config_t *config)
{
    enum {BUFSIZE = 40};
    char buf[BUFSIZE];
    uint32_t vers;

    buf[4] = 0;
    datain(f, buf, 4);
    memcpy(&config->ftyp,buf,4);
    vers = u32in(f);

    if (config->verbose.header)
        fprintf(stderr, "\nBrand:\t\t\t%s(version 0x%X)\n", buf, vers);

    stringin(f, buf, BUFSIZE);

    if (config->verbose.header)
        fprintf(stderr, "Compatible brands:\t%s\n", buf);

    return size;
}

/*
static int mvhdin(FILE *f, int size, mp4config_t *config)
{
    // version.8/flags.24
    u32in(f);
    // Creation time
    config->ctime = u32in(f);
    // Modification time
    config->mtime = u32in(f);
    // Time scale
    config->timescale = u32in(f);
    // Duration
    config->samples = u32in(f);
    // preferred playback rate (16.16)
    u32in(f);
    // preferred volume
    u16in(f);

    return size;
}
*/

static int mdhdin(FILE *f, int size, mp4config_t *config)
{
    // version/flags
    u32in(f);
    // Creation time
    config->ctime = u32in(f);
    // Modification time
    config->mtime = u32in(f);
    // Sampling time scale
    config->samplerate = u32in(f);
    // Duration
    config->samples = u32in(f);
    // Language
    u16in(f);
    // pre_defined
    u16in(f);

    return size;
};

static int hdlr1in(FILE *f, int size, mp4config_t *config)
{
    uint8_t buf[5];

    buf[4] = 0;
    // version/flags
    u32in(f);
    // pre_defined
    u32in(f);
    // Component subtype
    datain(f, buf, 4);
    if (config->verbose.header)
        fprintf(stderr, "*track media type: '%s': ", buf);
    if (memcmp("soun", buf, 4))
    {
        if (config->verbose.header)
            fprintf(stderr, "unsupported, skipping\n");
        return ERR_UNSUPPORTED;
    }
    else
    {
        if (config->verbose.header)
            fprintf(stderr, "OK\n");
    }
    // reserved
    u32in(f);
    u32in(f);
    u32in(f);
    // name
    // null terminate
    u8in(f);

    return size;
};

static int stsdin(FILE *f, int size, mp4config_t *config)
{
(void)config;
    // version/flags
    u32in(f);
    // Number of entries(one 'mp4a')
    if (u32in(f) != 1) //fixme: error handling
        return ERR_FAIL;

    return size;
};

static int mp4ain(FILE *f, int size, mp4config_t *config)
{
    uint16_t rateint, ratefrac;

    config->format = MAIN;
    // Reserved (6 bytes)
    u32in(f);
    u16in(f);
    // Data reference index
    u16in(f);
    // Version
    u16in(f);
    // Revision level
    u16in(f);
    // Vendor
    u32in(f);
    // Number of channels
    config->channels = u16in(f);
    // Sample size (bits)
    config->bits = u16in(f);
    // Compression ID
    u16in(f);
    // Packet size
    u16in(f);
    // Sample rate (16.16)
    // fractional framerate, probably not for audio
    // rate integer part
    rateint = u16in(f);
    // rate reminder part
    ratefrac = u16in(f);

    if (!config->samplerate)
    {
        config->samplerate = rateint;
        config->sampleratefrac = ratefrac;
    }

    return size;
}

static int ima4in(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0x13A4;
    return size;
}

static int alacin(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0xA1AC;
    return size;
}

static int alawin(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0xA1A3;
    return size;
}

static int ulawin(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0xB1A3;
    return size;
}

static int sowtin(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0x5034;
    return size;
}

static int twosin(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0x4305;
    return size;
}

static int MAC3in(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0x3AC3;
    return size;
}

static int MAC6in(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0x3AC6;
    return size;
}

static int in24in(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0x5124;
    return size;
}

static int in32in(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0x5132;
    return size;
}

static int in32lein(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0x1532;
    return size;
}

static int fl32in(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0xF132;
    return size;
}

static int fl64in(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0xF164;
    return size;
}

static int vorbin(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0x408B;
    return size;
}

static int opusin(FILE *f, int size, mp4config_t *config)
{
    if ((size=mp4ain(f,size,config)) >= 0)
        config->format = 0x0B25;
    return size;
}

int _mp4read_esds_atom(FILE *f, int size, mp4config_t *config)
{
    // descriptor tree:
    // MP4ES_Descriptor
    //   MP4DecoderConfigDescriptor
    //      MP4DecSpecificInfoDescriptor
    //   MP4SLConfigDescriptor
    enum{ TAG_MP4_ES = 3, TAG_DC = 4, TAG_DSI = 5, TAG_SLC = 6 };
    enum{ FMT_MP4 = 0x40, FMT_MP2_AAC_LC = 0x67, FMT_UNSPECIFIED = 0xFF};
    uint8_t format = 0, tag;

    tag = u8in(f);
    getsize(f);
    u16in(f); /* ID */
    config->buffersize = 0;
    if (tag == TAG_MP4_ES)
    {
        // flags(url(bit 6); ocr(5); streamPriority (0-4)):
        config->esflags[0] = u8in(f);

        if (u8in(f) != TAG_DC)
            return ERR_FAIL;
        getsize(f);
        format = u8in(f);
        // flags
        config->esflags[1] = u8in(f);
        // buffer size (24 bits)
        config->buffersize = u16in(f) << 8;
        config->buffersize |= u8in(f);
        // bitrate
        config->bitratemax = u32in(f);
        config->bitrateavg = u32in(f);
        // MP4-specific description tag
    }
#if 1
    else
        return ERR_FAIL;
#endif
    if (config->buffersize<128)
        config->buffersize = 32768; /* default large size */

    if (u8in(f) != TAG_DSI)
        return ERR_FAIL;
    config->asc.size = getsize(f);
    if (config->asc.size > sizeof(config->asc.buf))
        return ERR_FAIL;
    // get AudioSpecificConfig
    datain(f, config->asc.buf, config->asc.size);

    if (config->asc.size < 2)
        return ERR_FAIL;
    if (format != FMT_MP4)
        return format ? ERR_CODEC_UNSUPPORTED:ERR_FAIL; /* must be non-zero */
    tag = config->asc.buf[0] >> 3;
    if (tag == 31)
        tag = ((config->asc.buf[0] << 3 & 63) | config->asc.buf[1] >> 5) + 32;
    config->format = tag;
    switch(tag)
    {
    case MAIN: case LC: case SSR: case LTP: case HE_AAC: case SCALABLE:
    case ER_LC: case ER_LTP: case LD: case DRM_ER_LC: case SBR_PS:
    case 0:
        break;
    default:
        return ERR_CODEC_UNSUPPORTED;
    }

    if (u8in(f) != TAG_SLC)
        return ERR_FAIL;
    getsize(f);
    // "predefined" (no idea)
    u8in(f);

    return size;
}

int esdsin(FILE *f, int size, mp4config_t *config)
{
    // version/flags
    u32in(f);
    return _mp4read_esds_atom(f,size,config);
}

static int sttsin(FILE *f, int size, mp4config_t *config)
{
(void)f; (void)config;
    if (size < 16) //min stts size
        return ERR_FAIL;

    return size;
}

static int stszin(FILE *f, int size, mp4config_t *config)
{
    int cnt;
    uint32_t ofs, fsize;

    // version/flags
    u32in(f);
    // Sample size
    fsize = u32in(f);
    // Number of entries
    cnt = u32in(f);

    if (fsize)
    {
        config->samples = cnt / fsize;
        return size;
    }

    if (!(cnt + 1))
        return ERR_FAIL;
    config->frame.ents = cnt;

    config->frame.data = malloc(sizeof(*config->frame.data)
                                  * (config->frame.ents + 1));
    if (!config->frame.data)
        return ERR_MEMORY;

    ofs = 0;
    config->frame.data[0] = ofs;
    for (cnt = 0; cnt < config->frame.ents; cnt++)
    {
        fsize = u32in(f);
        ofs += fsize;
        if (config->frame.maxsize < fsize)
            config->frame.maxsize = fsize;

        config->frame.data[cnt + 1] = ofs;

        if (ofs < config->frame.data[cnt])
            return ERR_FAIL;
    }

    return size;
}

static int stcoin(FILE *f, int size, mp4config_t *config)
{
    uint32_t cnt;
    // version/flags
    u32in(f);
    // Number of entries
    cnt = u32in(f);
    if (cnt < 1)
        return ERR_FAIL;
    // first chunk offset
    config->dataofs = u32in(f);

    if (config->format > 255 && cnt > 1)
    { /* get first largest continguous block */
        uint32_t next, ofs, size1;

	ofs = u32in(f);
	size1 = ofs - config->dataofs;
	for (--cnt; --cnt; ofs = next)
	{
            next = u32in(f);
            if (next - ofs != size1)
                break;
	}
	config->datasize = ofs - config->dataofs;
    }

    // ignore the rest

    return size;
}

#if 0
static int tagtxt(char *tagname, const char *tagtxt)
{
    //int txtsize = strlen(tagtxt);
    int size = 0;
    //int datasize = txtsize + 16;

#if 0
    size += u32out(datasize + 8);
    size += dataout(tagname, 4);
    size += u32out(datasize);
    size += dataout("data", 4);
    size += u32out(1);
    size += u32out(0);
    size += dataout(tagtxt, txtsize);
#endif

    return size;
}

static int tagu32(char *tagname, int n /*number of stored fields*/)
{
    //int numsize = n * 4;
    int size = 0;
    //int datasize = numsize + 16;

#if 0
    size += u32out(datasize + 8);
    size += dataout(tagname, 4);
    size += u32out(datasize);
    size += dataout("data", 4);
    size += u32out(0);
    size += u32out(0);
#endif

    return size;
}
#endif

static int metain(FILE *f, int size)
{
    // version/flags
    u32in(f);

    return ERR_OK;
};

static int hdlr2in(FILE *f, int size)
{
    uint8_t buf[4];

    // version/flags
    u32in(f);
    // Predefined
    u32in(f);
    // Handler type
    datain(f, buf, 4);
    if (memcmp(buf, "mdir", 4))
        return ERR_FAIL;
    datain(f, buf, 4);
    if (memcmp(buf, "appl", 4))
        return ERR_FAIL;
    // Reserved
    u32in(f);
    u32in(f);
    // null terminator
    u8in(f);

    return size;
};

static int ilstin(FILE *f, int size)
{
    enum {NUMSET = 1, GENRE, EXTAG};
    int read = 0;

    static struct {
        char *name;
        char *id;
        int flag;
    } tags[] = {
        {"Album       ", "\xa9" "alb"},
        {"Album Artist", "aART"},
        {"Artist      ", "\xa9" "ART"},
        {"Comment     ", "\xa9" "cmt"},
        {"Cover image ", "covr"},
        {"Compilation ", "cpil"},
        {"Copyright   ", "cprt"},
        {"Date        ", "\xa9" "day"},
        {"Disc#       ", "disk", NUMSET},
        {"Genre       ", "gnre", GENRE},
        {"Grouping    ", "\xa9" "grp"},
        {"Lyrics      ", "\xa9" "lyr"},
        {"Title       ", "\xa9" "nam"},
        {"Rating      ", "rtng"},
        {"BPM         ", "tmpo"},
        {"Encoder     ", "\xa9" "too"},
        {"Track       ", "trkn", NUMSET},
        {"Composer    ", "\xa9" "wrt"},
        {0, "----", EXTAG},
        {0},
    };

    static const char *genres[] = {
        "Blues", "Classic Rock", "Country", "Dance",
        "Disco", "Funk", "Grunge", "Hip-Hop",
        "Jazz", "Metal", "New Age", "Oldies",
        "Other", "Pop", "R&B", "Rap",
        "Reggae", "Rock", "Techno", "Industrial",
        "Alternative", "Ska", "Death Metal", "Pranks",
        "Soundtrack", "Euro-Techno", "Ambient", "Trip-Hop",
        "Vocal", "Jazz+Funk", "Fusion", "Trance",
        "Classical", "Instrumental", "Acid", "House",
        "Game", "Sound Clip", "Gospel", "Noise",
        "Alternative Rock", "Bass", "Soul", "Punk",
        "Space", "Meditative", "Instrumental Pop", "Instrumental Rock",
        "Ethnic", "Gothic", "Darkwave", "Techno-Industrial",
        "Electronic", "Pop-Folk", "Eurodance", "Dream",
        "Southern Rock", "Comedy", "Cult", "Gangsta",
        "Top 40", "Christian Rap", "Pop/Funk", "Jungle",
        "Native US", "Cabaret", "New Wave", "Psychadelic",
        "Rave", "Showtunes", "Trailer", "Lo-Fi",
        "Tribal", "Acid Punk", "Acid Jazz", "Polka",
        "Retro", "Musical", "Rock & Roll", "Hard Rock",
        "Folk", "Folk-Rock", "National Folk", "Swing",
        "Fast Fusion", "Bebob", "Latin", "Revival",
        "Celtic", "Bluegrass", "Avantgarde", "Gothic Rock",
        "Progressive Rock", "Psychedelic Rock", "Symphonic Rock", "Slow Rock",
        "Big Band", "Chorus", "Easy Listening", "Acoustic",
        "Humour", "Speech", "Chanson", "Opera",
        "Chamber Music", "Sonata", "Symphony", "Booty Bass",
        "Primus", "Porn Groove", "Satire", "Slow Jam",
        "Club", "Tango", "Samba", "Folklore",
        "Ballad", "Power Ballad", "Rhythmic Soul", "Freestyle",
        "Duet", "Punk Rock", "Drum Solo", "Acapella",
        "Euro-House", "Dance Hall", "Goa", "Drum & Bass",
        "Club - House", "Hardcore", "Terror", "Indie",
        "BritPop", "Negerpunk", "Polsk Punk", "Beat",
        "Christian Gangsta Rap", "Heavy Metal", "Black Metal", "Crossover",
        "Contemporary Christian", "Christian Rock", "Merengue", "Salsa",
        "Thrash Metal", "Anime", "JPop", "Synthpop",
        "Unknown",
    };

    fprintf(stderr, "----------tag list-------------\n");
    while(read < size)
    {
        int asize, dsize;
        uint8_t id[5];
        int cnt;
        uint32_t type;

        id[4] = 0;

        asize = u32in(f);
        read += asize;
        asize -= 4;
        if (datain(f, id, 4) < 4)
            return ERR_FAIL;
        asize -= 4;

        for (cnt = 0; tags[cnt].id; cnt++)
        {
            if (!memcmp(id, tags[cnt].id, 4))
                break;
        }

        if (tags[cnt].name)
            fprintf(stderr, "%s :   ", tags[cnt].name);
        else
        {
            if (tags[cnt].flag != EXTAG)
                fprintf(stderr, "'%s'       :   ", id);
        }

        dsize = u32in(f);
        asize -= 4;
        if (datain(f, id, 4) < 4)
            return ERR_FAIL;
        asize -= 4;

        if (tags[cnt].flag != EXTAG)
        {
            if (memcmp(id, "data", 4))
                return ERR_FAIL;
        }
        else
        {
            int spc;

            if (memcmp(id, "mean", 4))
                goto skip;
            dsize -= 8;
            while (dsize > 0)
            {
                u8in(f);
                asize--;
                dsize--;
            }
            if (asize >= 8)
            {
                dsize = u32in(f) - 8;
                asize -= 4;
                if (datain(f, id, 4) < 4)
                    return ERR_FAIL;
                asize -= 4;
                if (memcmp(id, "name", 4))
                    goto skip;
                u32in(f);
                asize -= 4;
                dsize -= 4;
            }
            spc = 13 - dsize;
            if (spc < 0) spc = 0;
            while (dsize > 0)
            {
                fprintf(stderr, "%c",u8in(f));
                asize--;
                dsize--;
            }
            while (spc--)
                fprintf(stderr, " ");
            fprintf(stderr, ":   ");
            if (asize >= 8)
            {
                dsize = u32in(f) - 8;
                asize -= 4;
                if (datain(f, id, 4) < 4)
                    return ERR_FAIL;
                asize -= 4;
                if (memcmp(id, "data", 4))
                    goto skip;
                u32in(f);
                asize -= 4;
                dsize -= 4;
            }
            while (dsize > 0)
            {
                fprintf(stderr, "%c",u8in(f));
                asize--;
                dsize--;
            }
            fprintf(stderr, "\n");

            goto skip;
        }
        type = u32in(f);
        asize -= 4;
        u32in(f);
        asize -= 4;

        switch(type)
        {
        case 1:
            while (asize > 0)
            {
                fprintf(stderr, "%c",u8in(f));
                asize--;
            }
            break;
        case 0:
            switch(tags[cnt].flag)
            {
            case NUMSET:
                u16in(f);
                asize -= 2;

                fprintf(stderr, "%d", u16in(f));
                asize -= 2;
                fprintf(stderr, "/%d", u16in(f));
                asize -= 2;
                break;
            case GENRE:
                {
                    uint8_t gnum = u16in(f);
                    asize -= 2;
                    if (!gnum)
                       goto skip;
                    gnum--;
                    if (gnum >= 147)
                        gnum = 147;
                    fprintf(stderr, "%s", genres[gnum]);
                }
                break;
            default:
                while(asize > 0)
                {
                    fprintf(stderr, "%d/", u16in(f));
                    asize-=2;
                }
            }
            break;
        case 0x15:
            //fprintf(stderr, "(8bit data)");
            while(asize > 0)
            {
                fprintf(stderr, "%d", u8in(f));
                asize--;
                if (asize)
                    fprintf(stderr, "/");
            }
            break;
        case 0xd:
            fprintf(stderr, "(image data)");
            break;
        default:
            fprintf(stderr, "(unknown data type)");
            break;
        }
        fprintf(stderr, "\n");

    skip:
        // skip to the end of atom
        while (asize > 0)
        {
            u8in(f);
            asize--;
        }
    }
    fprintf(stderr, "-------------------------------\n");

    return size;
};

static int parse(FILE *f, uint32_t *sizemax,
        mp4config_t *config, creator_t **atom)
{
    long apos = 0;
    long aposmax = ftell(f) + *sizemax;
    uint32_t size, opc = (**atom).opcode;

    if (opc != ATOM_NAME && opc != ATOM_FORMAT && opc != ATOM_DESCRIPT)
    {
        fprintf(stderr, "mp4 parse error: root is not a 'name' opcode (%d)\n",
	    opc);
        return ERR_FAIL;
    }
    //fprintf(stderr, "looking for '%s'\n", (char *)g_atom->data);

    // search for atom in the file
    while (1)
    {
        char name[4];
        uint32_t tmp;

        apos = ftell(f);
        if (apos >= (aposmax - 8))
        {
            if (opc == ATOM_FORMAT || opc == ATOM_DESCRIPT)
            {
                (*atom)++;
                if ((**atom).opcode == ATOM_DATA)
                    (*atom)++;
                return ERR_OK;
            }
            if (config->verbose.errors)
                fprintf(stderr, "mp4 parse error: atom '%s' not found\n", (char *)(**atom).data);
            return ERR_FAIL;
        }
        if ((tmp = u32in(f)) < 8)
        {
            if (config->verbose.errors)
                fprintf(stderr, "mp4 parse error: invalid atom size %x @%lx\n", tmp, ftell(f));
            return ERR_FAIL;
        }

        size = tmp;
        if (datain(f, name, 4) != 4)
        { /* EOF / read error */
            if (config->verbose.errors)
                fprintf(stderr, "mp4 parse error: can't read atom name @%lx\n", ftell(f));
            return ERR_FAIL;
        }

        //fprintf(stderr, "atom: '%c%c%c%c'(%x)", name[0],name[1],name[2],name[3], size);

        if (!memcmp(name, (**atom).data, 4))
        {
            //fprintf(stderr, "OK\n");
            break;
        }
        //fprintf(stderr, "\n");

        fseek(f, apos + size, SEEK_SET);
    }
    *sizemax = size;
    (*atom)++;
    if ((**atom).opcode == ATOM_DATA)
    {
        int err = ((fileAtomIO)(**atom).data)(f, size - 8, config, atom);
        if (err < ERR_OK)
        {
            fseek(f, apos + size, SEEK_SET);
            return err;
        }
        (*atom)++;
    }
    while (opc == ATOM_FORMAT && (**atom).opcode != ATOM_DESCENT && (**atom).opcode)
        (*atom)++;
    if ((**atom).opcode == ATOM_DESCENT)
    {
        long apos = ftell(f);

        //fprintf(stderr, "descent\n");
        (*atom)++;
        while ((**atom).opcode != ATOM_STOP)
        {
            uint32_t subsize = size - 8;
            int ret;
            if ((**atom).opcode == ATOM_ASCENT)
            {
                (*atom)++;
                break;
            }
            fseek(f, apos, SEEK_SET);
            if ((ret = parse(f, &subsize, config, atom)) < 0)
                return ret;
        }
        //fprintf(stderr, "ascent\n");
    }

    fseek(f, apos + size, SEEK_SET);

    return ERR_OK;
}



static int moovin(FILE *f, int sizemax, mp4config_t *config,
        creator_t **atom)
{
    long apos = ftell(f);
    uint32_t atomsize;
    creator_t *old_atom = *atom;
    int err, ret = sizemax;

    static creator_t mvhd[] = {
        {ATOM_NAME, "mvhd"},
        {0}
    };
    static creator_t trak[] = {
        {ATOM_NAME, "trak"},
        {ATOM_DESCENT},
        {ATOM_NAME, "tkhd"},
        {ATOM_NAME, "mdia"},
        {ATOM_DESCENT},
        {ATOM_NAME, "mdhd"},
        {ATOM_DATA, mdhdin},
        {ATOM_NAME, "hdlr"},
        {ATOM_DATA, hdlr1in},
        {ATOM_NAME, "minf"},
        {ATOM_DESCENT},
        {ATOM_NAME, "smhd"},
        {ATOM_NAME, "dinf"},
        {ATOM_NAME, "stbl"},
        {ATOM_DESCENT},
        {ATOM_NAME, "stsd"},
        {ATOM_DATA, stsdin},
        {ATOM_DESCENT},
        {ATOM_FORMAT, "mp4a"},
        {ATOM_DATA, mp4ain},
        {ATOM_FORMAT, "alac"},
        {ATOM_DATA, alacin},
        {ATOM_FORMAT, "ima4"},
        {ATOM_DATA, ima4in},
        {ATOM_FORMAT, "alaw"},
        {ATOM_DATA, alawin},
        {ATOM_FORMAT, "ulaw"},
        {ATOM_DATA, ulawin},
        {ATOM_FORMAT, "twos"},
        {ATOM_DATA, twosin},
        {ATOM_FORMAT, "sowt"},
        {ATOM_DATA, sowtin},
        {ATOM_FORMAT, "MAC3"},
        {ATOM_DATA, MAC3in},
        {ATOM_FORMAT, "MAC6"},
        {ATOM_DATA, MAC6in},
        {ATOM_FORMAT, "in24"},
        {ATOM_DATA, in24in},
        {ATOM_FORMAT, "in32"},
        {ATOM_DATA, in32in},
        {ATOM_FORMAT, "23ni"},
        {ATOM_DATA, in32lein},
        {ATOM_FORMAT, "fl32"},
        {ATOM_DATA, fl32in},
        {ATOM_FORMAT, "fl64"},
        {ATOM_DATA, fl64in},
        {ATOM_FORMAT, "vorb"},
        {ATOM_DATA, vorbin},
        {ATOM_FORMAT, "opus"},
        {ATOM_DATA, opusin},
        {ATOM_DESCENT},
        {ATOM_DESCRIPT, "esds"},
        {ATOM_DATA, esdsin},
        {ATOM_ASCENT},
        {ATOM_ASCENT},
        {ATOM_NAME, "stts"},
        {ATOM_DATA, sttsin},
        {ATOM_NAME, "stsc"},
        {ATOM_NAME, "stsz"},
        {ATOM_DATA, stszin},
        {ATOM_NAME, "stco"},
        {ATOM_DATA, stcoin},
        {0}
    };

    *atom = mvhd;
    atomsize = sizemax + apos - ftell(f);
    if (parse(f, &atomsize, config, atom) < 0) {
        *atom = old_atom;
        return ERR_FAIL;
    }

    fseek(f, apos, SEEK_SET);

    while (1)
    {
        //fprintf(stderr, "TRAK\n");
        *atom = trak;
        atomsize = sizemax + apos - ftell(f);
        if (atomsize < 8)
            break;
        //fprintf(stderr, "PARSE(%x)\n", atomsize);
        err = parse(f, &atomsize, config, atom);
        //fprintf(stderr, "SIZE: %x/%x\n", atomsize, sizemax);
        if (err >= 0)
            break;
        if (err != ERR_UNSUPPORTED) {
            ret = err;
            break;
        }
        //fprintf(stderr, "UNSUPP\n");
    }

    *atom = old_atom;
    return ret;
}


/* must have one or the other */
static creator_t g_head[] = {
    {ATOM_FORMAT, "ftyp"}, /* QT movie by default */
    {ATOM_DATA, ftypin},
    {0}
};
static creator_t g_mdat[] = {
    {ATOM_FORMAT, "mdat"},
    {ATOM_DATA, mdatin},
    {0}
};

static creator_t g_moov[] = {
    {ATOM_NAME, "moov"},
    {ATOM_DATA, moovin},
    //{ATOM_DESCENT},
    //{ATOM_NAME, "mvhd"},
    {0}
};

static creator_t g_meta1[] = {
    {ATOM_NAME, "moov"},
    {ATOM_DESCENT},
    {ATOM_NAME, "udta"},
    {ATOM_DESCENT},
    {ATOM_NAME, "meta"},
    {ATOM_DATA, metain},
    {ATOM_DESCENT},
    {ATOM_NAME, "hdlr"},
    {ATOM_DATA, hdlr2in},
    {ATOM_NAME, "ilst"},
    {ATOM_DATA, ilstin},
    {0}
};

static creator_t g_meta2[] = {
    {ATOM_NAME, "meta"},
    {ATOM_DATA, metain},
    {ATOM_DESCENT},
    {ATOM_NAME, "hdlr"},
    {ATOM_DATA, hdlr2in},
    {ATOM_NAME, "ilst"},
    {ATOM_DATA, ilstin},
    {0}
};


int mp4read_frame(FILE *f, mp4config_t *config)
{
    if (config->frame.current >= config->frame.ents)
        return ERR_FAIL;

    config->bitbuf.size = config->frame.data[config->frame.current + 1]
        - config->frame.data[config->frame.current];

    if (config->bitbuf.size > config->buffersize
        || fread(config->bitbuf.data, 1, config->bitbuf.size, f)
        != config->bitbuf.size)
    {
        if (config->verbose.errors)
            fprintf(stderr, "can't read frame data(frame %d@0x%x)\n",
                config->frame.current,
                config->frame.data[config->frame.current]);

        return ERR_FAIL;
    }

    config->frame.current++;

    return ERR_OK;
}

int mp4read_seek(FILE *f, mp4config_t *config, int framenum)
{
    if (framenum > config->frame.ents)
        return ERR_FAIL;
    if (fseek(f, config->dataofs + (config->frame.data ? config->frame.data[framenum]
        : framenum*config->frame.maxsize), SEEK_SET))
        return ERR_FAIL;

    config->frame.current = framenum;

    return ERR_OK;
}

#ifndef NO_MP4_HEADER_INFO
static void mp4info(const mp4config_t *config)
{
    fprintf(stderr, "Modification Time:\t\t%s\n", mp4time(config->mtime));
    fprintf(stderr, "Samplerate:\t\t%d\n", config->samplerate);
    fprintf(stderr, "Total samples:\t\t%d\n", config->samples);
    fprintf(stderr, "Total channels:\t\t%d\n", config->channels);
    fprintf(stderr, "Bits per sample:\t%d\n", config->bits);
    fprintf(stderr, "Buffer size:\t\t%d\n", config->buffersize);
    fprintf(stderr, "Max bitrate:\t\t%d\n", config->bitratemax);
    fprintf(stderr, "Average bitrate:\t%d\n", config->bitrateavg);
//  fprintf(stderr, "Samples per frame:\t%d\n", config->framesamples);
    fprintf(stderr, "Frames:\t\t\t%d\n", config->frame.ents);
    fprintf(stderr, "ASC size:\t\t%d\n", config->asc.size);
    if (config->samplerate)
        fprintf(stderr, "Duration:\t\t%.1f sec\n",(float)config->samples/config->samplerate);
    fprintf(stderr, "Data offset/size:\t%x/%x\n", config->dataofs, config->datasize);
}
#endif

int mp4read_close(FILE *f, mp4config_t *config)
{
#define FREE(x) if(x){free(x);x=0;}
    FREE(config->frame.data);
    FREE(config->bitbuf.data);

    if (f)
        fclose(f);

    return ERR_OK;
}


int _mp4read_open(FILE *f, mp4config_t *config)
{
    creator_t *atom;
    uint32_t atomsize, filesize, initoffs = ftell(f);
    int ret;

    fseek(f,0,SEEK_END);
    filesize = ftell(f);
    fseek(f,initoffs,SEEK_SET);
#if 0 /* not guaranteed to be a file we can recognize */
    if (config->verbose.header)
        fprintf(stderr, "**** MP4 header ****\n");
#endif
    atom = g_head;
    atomsize = filesize; //INT_MAX;
    if ((ret = parse(f, &atomsize, config, &atom)) < 0)
        goto perr;
    if (!config->ftyp)
    {
        fseek(f,initoffs,SEEK_SET);
        atom = g_mdat;
        atomsize = filesize; //INT_MAX;
        if ((ret = parse(f, &atomsize, config, &atom)) < 0)
            goto perr;
        ret = ERR_FAIL;
        if (!config->ftyp)
            goto perr;
    }
    atom = g_moov;
    atomsize = filesize; //INT_MAX;
    fseek(f,initoffs,SEEK_SET);
    if ((ret = parse(f, &atomsize, config, &atom)) < 0)
    {
    perr:
        if (ret == ERR_CODEC_UNSUPPORTED)
        {
            ret = ERR_CODEC_UNSUPPORTED;
            goto err;
        }
        if (ret != ERR_FAIL && config->verbose.errors)
            fprintf(stderr, "mp4read_open->parse error:%d\n", ret);
        goto ferr;
    }

    if (config->format > 255)
    {
        ret = ERR_CODEC_UNSUPPORTED;
        goto err;
    }
    if (!config->format)
    {
    ferr:
        ret = ERR_FAIL;
        goto err;
    }
    // alloc frame buffer
    config->bitbuf.data = malloc(config->frame.maxsize);
    if (!config->bitbuf.data)
    {
        ret = ERR_MEMORY;
    err:
        if (config->frame.data)
	{
            free(config->frame.data);
            config->frame.data = NULL;
        }
        return ret;
    }

#ifndef NO_MP4_HEADER_INFO
    if (config->verbose.header)
    {
        mp4info(config);
        fprintf(stderr, "********************\n");
    }
#endif
    if (config->verbose.tags)
    {
        rewind(f);
        atom = g_meta1;
        atomsize = INT_MAX;
        ret = parse(f, &atomsize, config, &atom);
        if (ret < 0)
        {
            rewind(f);
            atom = g_meta2;
            atomsize = INT_MAX;
            ret = parse(f, &atomsize, config, &atom);
        }
    }
    return ERR_OK;
}

FILE* mp4read_open(char *name, mp4config_t *config, int *err)
{
    FILE *f;
    int ret = ERR_FAIL;

    mp4read_close(NULL,config);

    f = faad_fopen(name, "rb");
    if (!f)
        goto ffail;
    ret = _mp4read_open(f,config);
    if (!ret)
    {
        *err = ERR_OK;
        return f;
    }
    mp4read_close(f, config);
ffail:
    if (err)
        *err = ret;
    return NULL;
}
